//#define _GNU_SOURCE
#include "ga.h"
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <wait.h>
#include <fcntl.h>

void Ga::run(){
    params = new Params;
    params->evaluator = this;
    loadParameters();
    buildGenome();
    buildPopulation();
    done = false;
    showParse = false;
    for( int i = 0; i < params->numGens; i ++ ){
        pop.runGeneration();
        showBest();
    }
    pop.logGenStats();
}
    
void Ga::loadParameters(){
    params->bitsPerFloat = 16;
    params->popSize = 10;
    params->numGens = 10;
    params->mutationProb = .01;
    params->crossoverProb = .5;
    params->elitists = 0;

    FILE* file = fopen("parameters", "r");
    if( file <= 0 ) return;
    char param[1024];
    float value;
    int i;
    while( i = fscanf( file, "%s %f\n", &param, &value ) == 2 ){
        printf( "reading in %s=%f\n", param, value );
             if( strcmp( param, "popsize" )         == 0) params->popSize = (int) value;
        else if( strcmp( param, "numgens" )         == 0) params->numGens = (int) value;
        else if( strcmp( param, "mutationprob" )    == 0) params->mutationProb = value;
        else if( strcmp( param, "crossoverprob" )   == 0) params->crossoverProb = value;
        else if( strcmp( param, "elitists" )        == 0) params->elitists = (int) value;
        else if( strcmp( param, "bitsperfloat" )    == 0) params->bitsPerFloat = (int) value;
        else if( strcmp( param, "problem" )         == 0) problem = (int) value;
    }
    fclose( file );
    printf("%s\n", params->toString() );
}

void Ga::buildGenome(){
    repres.setParams( params );
    Chrom c;
    c.setParams( params );
    if( problem == 1 ){
        c.setType( c.FLOAT );
        c.setMin( -5.12f );
        c.setMax(  5.12f );
        repres.add( c, "x1" );
        repres.add( c, "x2" );
        repres.add( c, "x3" );
    }
    if( problem == 2 ){
        c.setType( c.FLOAT );
        c.setMin( -2.048f );
        c.setMax(  2.048f );
        repres.add( c, "x1" );
        repres.add( c, "x2" );
    }
    if( problem == 3 ){
        c.setType( c.FLOAT );
        c.setMin( -5.12f );
        c.setMax(  5.12f );
        repres.add( c, "x1" );
        repres.add( c, "x2" );
        repres.add( c, "x3" );
        repres.add( c, "x4" );
        repres.add( c, "x5" );
    }
    if( problem == 4 ){
        c.setType( c.FLOAT );
        c.setMin( -1.28f );
        c.setMax(  1.28f );
        for( int i = 0; i < 30; i ++ )
            repres.add( c, "" ); //not going to use handles
    }
    if( problem == 5 ){
        c.setType( c.FLOAT );
        c.setMin( -65.536 );
        c.setMin(  65.536 );
        for( int i = 0; i < 25; i ++ )
            repres.add( c, "" );
    }

    params->bitTree = new BitTree( params->genomeLength );
}

void Ga::buildPopulation(){
    pop.setParams( params );
    pop.initialize();
    printf("%s\n", pop.toString() );
}


#define sqr(num) ((num)*(num))
float gauss( float mean, float stddev ){
    float x1, x2, w, y1;
    static float y2;
    bool odd = false;
    if( odd ) {
        odd = false;
        return y2;
    }
    //polar box-muller transform -> gaussian numbers (faster then central limit method)
    do {
        x1 = 2.0f * drand48() - 1.0f;
        x2 = 2.0f * drand48() - 1.0f;
        w = sqr(x1) + sqr(x2);
    } while ( w >= 1.0f );
    w = sqrt( ( -2.0f * log( w ) )  / w );
    y1 = x1 * w;
    y1 = y1 * stddev + mean;
    //y2 is another gaussian number
    y2 = x2 * w;
    y2 = y2 * stddev + mean;
    odd = true;
    return y1;
}

float Ga::eval( Genome* g ){
    float fitness = 0;
    if( showParse )
        printf("%s\n", g->toString().Get() );

    repres.startParse( g );
    if( problem == 1 ){
        static RepresHandle hx1 = repres.getHandle("x1");
        static RepresHandle hx2 = repres.getHandle("x2");
        static RepresHandle hx3 = repres.getHandle("x3");
        float x1 = repres.parseF( hx1 );
        float x2 = repres.parseF( hx2 );
        float x3 = repres.parseF( hx3 );
        fitness = sqr(x1) + sqr(x2) + sqr(x3);
        //fitness = sqr(5.12f) * 3.0f - fabs( fitness );
        fitness *= -1;
        if( showParse )
            printf("x1:%f x2:%f x3:%f fit : %f\n", x1, x2, x3, fitness );
    }
    if( problem == 2 ){
        static RepresHandle hx1 = repres.getHandle("x1");
        static RepresHandle hx2 = repres.getHandle("x2");
        float x1 = repres.parseF( hx1 );
        float x2 = repres.parseF( hx2 );
        fitness = 100.0f * sqr(sqr(x1) - x2) + sqr(1.0f - x1);
        //fitness = 3900.0f - fitness;
        fitness *= -1;
        if( showParse )
            printf("x1:%f x2:%f fit:%f\n", x1, x2, fitness );
    }
    if( problem == 3 ){
        static RepresHandle hx1 = repres.getHandle("x1");
        static RepresHandle hx2 = repres.getHandle("x2");
        static RepresHandle hx3 = repres.getHandle("x3");
        static RepresHandle hx4 = repres.getHandle("x4");
        static RepresHandle hx5 = repres.getHandle("x5");
        float x1 = repres.parseF( hx1 );
        float x2 = repres.parseF( hx2 );
        float x3 = repres.parseF( hx3 );
        float x4 = repres.parseF( hx4 );
        float x5 = repres.parseF( hx5 );
        fitness = ((int) x1) + ((int)x2) + ((int)x3) + ((int)x4) + ((int)x5);
        //fitness = 5 * 5 - fitness;
        fitness *= -1;
        if( showParse )
            printf("x1:%f x2:%f x3:%f x4:%f x5:%f fit:%f\n", x1, x2, x3, x4, x5, fitness );
    }
    if( problem == 4 ){
            
        for( int i = 0; i < 30; i ++ ){
            float x = repres.parseF( i );
            fitness += (i+1) * sqr(sqr(x));
        }
        fitness += gauss(0,1);
        fitness *= -1;
    }
    if( problem == 5 ){
    }

    repres.stopParse();
    return fitness;
}

void Ga::showBest(){
    showParse = true;
    eval ( &pop[0] );
    showParse = false;
}
